@@ -14,7 +14,7 @@ static unsigned char * DirWithThumbnailPtrs;
 static double FocalplaneXRes;
 static double FocalplaneUnits;
 static int ExifImageWidth;
-static int MotorolaOrder = 0;
+int MotorolaOrder = 0;
 
 // for fixing the rotation.
 static void * OrientationPtr[2];
@@ -143,6 +143,7 @@ const int BytesPerFormat[] = {0,1,1,2,4,8,1,1,2,4,8,4,8};
 #define TAG_LIGHT_SOURCE           0x9208
 #define TAG_FLASH                  0x9209
 #define TAG_FOCALLENGTH            0x920A
+#define TAG_SUBJECTAREA            0x9214
 #define TAG_MAKER_NOTE             0x927C
 #define TAG_USERCOMMENT            0x9286
 #define TAG_SUBSEC_TIME            0x9290
@@ -157,8 +158,8 @@ const int BytesPerFormat[] = {0,1,1,2,4,8,1,1,2,4,8,4,8};
 
 #define TAG_FLASH_PIX_VERSION      0xA000
 #define TAG_COLOR_SPACE            0xA001
-#define TAG_EXIF_IMAGEWIDTH        0xA002
-#define TAG_EXIF_IMAGELENGTH       0xA003
+#define TAG_PIXEL_X_DIMENSION      0xA002
+#define TAG_PIXEL_Y_DIMENSION      0xA003
 #define TAG_RELATED_AUDIO_FILE     0xA004
 #define TAG_INTEROP_OFFSET         0xA005
 #define TAG_FLASH_ENERGY           0xA20B
@@ -183,6 +184,7 @@ const int BytesPerFormat[] = {0,1,1,2,4,8,1,1,2,4,8,4,8};
 #define TAG_SATURATION             0xA409
 #define TAG_SHARPNESS              0xA40A
 #define TAG_DISTANCE_RANGE         0xA40C
+#define TAG_IMAGE_UNIQUE_ID        0xA420
 
 static const TagTable_t TagTable[] = {
   { TAG_INTEROP_INDEX,          "InteropIndex"},
@@ -263,8 +265,8 @@ static const TagTable_t TagTable[] = {
   { TAG_WINXP_SUBJECT,          "Windows-XP subject"},
   { TAG_FLASH_PIX_VERSION,      "FlashPixVersion"},
   { TAG_COLOR_SPACE,            "ColorSpace"},
-  { TAG_EXIF_IMAGEWIDTH,        "ExifImageWidth"},
-  { TAG_EXIF_IMAGELENGTH,       "ExifImageLength"},
+  { TAG_PIXEL_X_DIMENSION,      "ExifImageWidth"},
+  { TAG_PIXEL_Y_DIMENSION,      "ExifImageLength"},
   { TAG_RELATED_AUDIO_FILE,     "RelatedAudioFile"},
   { TAG_INTEROP_OFFSET,         "InteroperabilityOffset"},
   { TAG_FLASH_ENERGY,           "FlashEnergy"},              
@@ -283,12 +285,14 @@ static const TagTable_t TagTable[] = {
   { TAG_WHITEBALANCE,           "WhiteBalance"},
   { TAG_DIGITALZOOMRATIO,       "DigitalZoomRatio"},
   { TAG_FOCALLENGTH_35MM,       "FocalLengthIn35mmFilm"},
+  { TAG_SUBJECTAREA,            "SubjectArea"},
   { TAG_SCENE_CAPTURE_TYPE,     "SceneCaptureType"},
   { TAG_GAIN_CONTROL,           "GainControl"},
   { TAG_CONTRAST,               "Contrast"},
   { TAG_SATURATION,             "Saturation"},
   { TAG_SHARPNESS,              "Sharpness"},
   { TAG_DISTANCE_RANGE,         "SubjectDistanceRange"},
+  { TAG_IMAGE_UNIQUE_ID,        "ImageUniqueId"},
 } ;
 
 #define TAG_TABLE_SIZE  (sizeof(TagTable) / sizeof(TagTable_t))
@@ -376,10 +380,14 @@ void PrintFormatNumber(void * ValuePtr, int Format, int ByteCount)
             case FMT_SLONG:     printf("%d",Get32s(ValuePtr)); s=4;      break;
             case FMT_SSHORT:    printf("%hd",(signed short)Get16u(ValuePtr)); s=2; break;
             case FMT_URATIONAL:
+                printf("%u/%u",Get32s(ValuePtr), Get32s(4+(char *)ValuePtr)); 
+                s = 8;
+                break;
+
             case FMT_SRATIONAL: 
-               printf("%d/%d",Get32s(ValuePtr), Get32s(4+(char *)ValuePtr)); 
-               s = 8;
-               break;
+                printf("%d/%d",Get32s(ValuePtr), Get32s(4+(char *)ValuePtr)); 
+                s = 8;
+                break;
 
             case FMT_SINGLE:    printf("%f",(double)*(float *)ValuePtr); s=8; break;
             case FMT_DOUBLE:    printf("%f",*(double *)ValuePtr);        s=8; break;
@@ -421,7 +429,11 @@ double ConvertAnyFormat(void * ValuePtr, int Format)
                 if (Den == 0){
                     Value = 0;
                 }else{
-                    Value = (double)Num/Den;
+                    if (Format == FMT_SRATIONAL){
+                        Value = (double)Num/Den;
+                    }else{
+                        Value = (double)(unsigned)Num/(double)(unsigned)Den;
+                    }
                 }
                 break;
             }
@@ -477,7 +489,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
             }
         }
         if (DumpExifMap){
-            printf("Map: %05d-%05d: Directory\n",(int)(DirStart-OffsetBase), (int)(DirEnd+4-OffsetBase));
+            printf("Map: %05u-%05u: Directory\n",(int)(DirStart-OffsetBase), (int)(DirEnd+4-OffsetBase));
         }
 
 
@@ -527,7 +539,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
             }
 
             if (DumpExifMap){
-                printf("Map: %05d-%05d:   Data for tag %04x\n",OffsetVal, OffsetVal+ByteCount, Tag);
+                printf("Map: %05u-%05u:   Data for tag %04x\n",OffsetVal, OffsetVal+ByteCount, Tag);
             }
         }else{
             // 4 bytes or less and value is in the dir entry itself
@@ -546,13 +558,11 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
             // Show tag name
             for (a=0;;a++){
                 if (a >= TAG_TABLE_SIZE){
-                    printf(IndentString);
-                    printf("    Unknown Tag %04x Value = ", Tag);
+                    printf("%s    Unknown Tag %04x Value = ", IndentString, Tag);
                     break;
                 }
                 if (TagTable[a].Tag == Tag){
-                    printf(IndentString);
-                    printf("    %s = ",TagTable[a].Desc);
+                    printf("%s    %s = ",IndentString, TagTable[a].Desc);
                     break;
                 }
             }
@@ -642,7 +652,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 if (ByteCount > 1){
                     if (ByteCount > MAX_COMMENT_SIZE) ByteCount = MAX_COMMENT_SIZE;
                     memcpy(ImageInfo.Comments, ValuePtr, ByteCount);
-                    ImageInfo.CommentWidchars = ByteCount/2;
+                    ImageInfo.CommentWidthchars = ByteCount/2;
                 }
                 break;
 
@@ -665,17 +675,21 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 }
 
                 // Copy the comment
-                if (memcmp(ValuePtr, "ASCII",5) == 0){
-                    for (a=5;a<10;a++){
-                        int c;
-                        c = (ValuePtr)[a];
-                        if (c != '\0' && c != ' '){
-                            strncpy(ImageInfo.Comments, (char *)ValuePtr+a, 199);
-                            break;
+                {
+                    int msiz = ExifLength - (ValuePtr-OffsetBase);
+                    if (msiz > ByteCount) msiz = ByteCount;
+                    if (msiz > MAX_COMMENT_SIZE-1) msiz = MAX_COMMENT_SIZE-1;
+                    if (msiz > 5 && memcmp(ValuePtr, "ASCII",5) == 0){
+                        for (a=5;a<10 && a < msiz;a++){
+                            int c = (ValuePtr)[a];
+                            if (c != '\0' && c != ' '){
+                                strncpy(ImageInfo.Comments, (char *)ValuePtr+a, msiz-a);
+                                break;
+                            }
                         }
+                    }else{
+                        strncpy(ImageInfo.Comments, (char *)ValuePtr, msiz);
                     }
-                }else{
-                    strncpy(ImageInfo.Comments, (char *)ValuePtr, MAX_COMMENT_SIZE-1);
                 }
                 break;
 
@@ -730,7 +744,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
             case TAG_ORIENTATION:
                 if (NumOrientations >= 2){
                     // Can have another orientation tag for the thumbnail, but if there's
-                    // a third one, things are stringae.
+                    // a third one, things are strange.
                     ErrNonfatal("More than two orientation in Exif",0,0);
                     break;
                 }
@@ -741,13 +755,12 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 }
                 if (ImageInfo.Orientation < 0 || ImageInfo.Orientation > 8){
                     ErrNonfatal("Undefined rotation value %d in Exif", ImageInfo.Orientation, 0);
-                    ImageInfo.Orientation = 0;
                 }
                 NumOrientations += 1;
                 break;
 
-            case TAG_EXIF_IMAGELENGTH:
-            case TAG_EXIF_IMAGEWIDTH:
+            case TAG_PIXEL_Y_DIMENSION:
+            case TAG_PIXEL_X_DIMENSION:
                 // Use largest of height and width to deal with images that have been
                 // rotated to portrait format.
                 a = (int)ConvertAnyFormat(ValuePtr, Format);
@@ -809,10 +822,6 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
 
             case TAG_ISO_EQUIVALENT:
                 ImageInfo.ISOequivalent = (int)ConvertAnyFormat(ValuePtr, Format);
-                if ( ImageInfo.ISOequivalent < 50 ){
-                    // Fixes strange encoding on some older digicams.
-                    ImageInfo.ISOequivalent *= 200;
-                }
                 break;
 
             case TAG_DIGITALZOOMRATIO:
@@ -854,7 +863,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                     if (SubdirStart < OffsetBase || SubdirStart > OffsetBase+ExifLength){
                         ErrNonfatal("Illegal GPS directory link in Exif",0,0);
                     }else{
-                        ProcessGpsInfo(SubdirStart, ByteCount, OffsetBase, ExifLength);
+                        ProcessGpsInfo(SubdirStart, OffsetBase, ExifLength);
                     }
                     continue;
                 }
@@ -872,6 +881,31 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
                 //   1 = macro, 2 = close, 3 = distant
                 ImageInfo.DistanceRange = (int)ConvertAnyFormat(ValuePtr, Format);
                 break;
+
+
+
+            case TAG_X_RESOLUTION:
+                if (NestingLevel==0) {// Only use the values from the top level directory
+                    ImageInfo.xResolution = (float)ConvertAnyFormat(ValuePtr,Format);
+                    // if yResolution has not been set, use the value of xResolution
+                    if (ImageInfo.yResolution == 0.0) ImageInfo.yResolution = ImageInfo.xResolution;
+                }
+                break;
+
+            case TAG_Y_RESOLUTION:
+                if (NestingLevel==0) {// Only use the values from the top level directory
+                    ImageInfo.yResolution = (float)ConvertAnyFormat(ValuePtr,Format);
+                    // if xResolution has not been set, use the value of yResolution
+                    if (ImageInfo.xResolution == 0.0) ImageInfo.xResolution = ImageInfo.yResolution;
+                }
+                break;
+
+            case TAG_RESOLUTION_UNIT:
+                if (NestingLevel==0) {// Only use the values from the top level directory
+                    ImageInfo.ResolutionUnit = (int) ConvertAnyFormat(ValuePtr,Format);
+                }
+                break;
+
         }
     }
 
@@ -932,7 +966,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
             ImageInfo.ThumbnailSize = ThumbnailSize;
 
             if (ShowTags){
-                printf("Thumbnail size: %d bytes\n",ThumbnailSize);
+                printf("Thumbnail size: %u bytes\n",ThumbnailSize);
             }
         }
     }
@@ -945,7 +979,7 @@ static void ProcessExifDir(unsigned char * DirStart, unsigned char * OffsetBase,
 //--------------------------------------------------------------------------
 void process_EXIF (unsigned char * ExifSection, unsigned int length)
 {
-    int FirstOffset;
+    unsigned int FirstOffset;
 
     FocalplaneXRes = 0;
     FocalplaneUnits = 0;
@@ -953,7 +987,7 @@ void process_EXIF (unsigned char * ExifSection, unsigned int length)
     NumOrientations = 0;
 
     if (ShowTags){
-        printf("Exif header %d bytes long\n",length);
+        printf("Exif header %u bytes long\n",length);
     }
 
     {   // Check the EXIF header component
@@ -985,6 +1019,10 @@ void process_EXIF (unsigned char * ExifSection, unsigned int length)
 
     FirstOffset = Get32u(ExifSection+12);
     if (FirstOffset < 8 || FirstOffset > 16){
+        if (FirstOffset < 16 || FirstOffset > length-16){
+            ErrNonfatal("invalid offset for first Exif IFD value",0,0);
+            return;
+        }
         // Usually set to 8, but other values valid too.
         ErrNonfatal("Suspicious offset of first Exif IFD value",0,0);
     }
@@ -1009,7 +1047,7 @@ void process_EXIF (unsigned char * ExifSection, unsigned int length)
 
 
     // Compute the CCD width, in millimeters.
-    if (FocalplaneXRes != 0){
+    if (FocalplaneXRes != 0 && ExifImageWidth != 0){
         // Note: With some cameras, its not possible to compute this correctly because
         // they don't adjust the indicated focal plane resolution units when using less
         // than maximum resolution, so the CCDWidth value comes out too small.  Nothing
@@ -1185,7 +1223,7 @@ const char * ClearOrientation(void)
 
             case FMT_ULONG:     
             case FMT_SLONG:     
-                memset(OrientationPtr, 0, 4);
+                memset(OrientationPtr[a], 0, 4);
                 // Can't be bothered to write  generic Put32 if I only use it once.
                 if (MotorolaOrder){
                     ((uchar *)OrientationPtr[a])[3] = 1;
@@ -1199,55 +1237,13 @@ const char * ClearOrientation(void)
         }
     }
 
-    return OrientTab[ImageInfo.Orientation];
+    if (ImageInfo.Orientation >= 1 && ImageInfo.Orientation <= 8){
+        return OrientTab[ImageInfo.Orientation];
+    }else{
+        return NULL;
+    }
 }
 
-
-
-//--------------------------------------------------------------------------
-// Remove thumbnail out of the exif image.
-//--------------------------------------------------------------------------
-int RemoveThumbnail(unsigned char * ExifSection)
-{
-    if (!DirWithThumbnailPtrs || 
-        ImageInfo.ThumbnailOffset == 0 || 
-        ImageInfo.ThumbnailSize == 0){
-        // No thumbnail, or already deleted it.
-        return 0;
-    }
-    if (ImageInfo.ThumbnailAtEnd == FALSE){
-        ErrNonfatal("Thumbnail not at end of Exif header, can't remove it", 0, 0);
-        return 0;
-    }
-
-    {
-        int de;
-        int NumDirEntries;
-        NumDirEntries = Get16u(DirWithThumbnailPtrs);
-
-        for (de=0;de<NumDirEntries;de++){
-            int Tag;
-            unsigned char * DirEntry;
-            DirEntry = DIR_ENTRY_ADDR(DirWithThumbnailPtrs, de);
-            Tag = Get16u(DirEntry);
-            if (Tag == TAG_THUMBNAIL_LENGTH){
-                // Set length to zero.
-                if (Get16u(DirEntry+2) != FMT_ULONG){
-                    // non standard format encoding.  Can't do it.
-                    ErrNonfatal("Can't remove Exif thumbnail", 0, 0);
-                    return 0;
-                }
-                Put32u(DirEntry+8, 0);
-            }                    
-        }
-    }
-
-    // This is how far the non thumbnail data went.
-    return ImageInfo.ThumbnailOffset+8;
-
-}
-
-
 //--------------------------------------------------------------------------
 // Convert exif time to Unix time structure
 //--------------------------------------------------------------------------
@@ -1266,6 +1262,17 @@ int Exif2tm(struct tm * timeptr, char * ExifTime)
             &timeptr->tm_hour, &timeptr->tm_min, &timeptr->tm_sec);
 
     if (a >= 5){
+        if (timeptr->tm_year <= 12 && timeptr->tm_mday > 2000 && ExifTime[2] == '.'){
+            // LG Electronics VX-9700 seems to encode the date as 'MM.DD.YYYY HH:MM'
+            // can't these people read the standard?  At least they left enough room
+            // in the header to put an Exif format date in there.
+            int tmp;
+            tmp = timeptr->tm_year;
+            timeptr->tm_year = timeptr->tm_mday;
+            timeptr->tm_mday = timeptr->tm_mon;
+            timeptr->tm_mon = tmp;
+        }
+
         // Accept five or six parameters.  Some cameras do not store seconds.
         timeptr->tm_isdst = -1;  
         timeptr->tm_mon -= 1;      // Adjust for unix zero-based months 
@@ -1303,7 +1310,7 @@ void ShowImageInfo(int ShowFileInfo)
     }
     printf("Resolution   : %d x %d\n",ImageInfo.Width, ImageInfo.Height);
 
-    if (ImageInfo.Orientation > 1){
+    if (ImageInfo.Orientation > 1 && ImageInfo.Orientation <=8){
         // Only print orientation if one was supplied, and if its not 1 (normal orientation)
         printf("Orientation  : %s\n", OrientTab[ImageInfo.Orientation]);
     }
@@ -1427,17 +1434,17 @@ void ShowImageInfo(int ShowFileInfo)
             // don't bother showing it - it doesn't add any useful information.
     }
 
-    if (ImageInfo.MeteringMode){ // 05-jan-2001 vcs
+    if (ImageInfo.MeteringMode > 0){ // 05-jan-2001 vcs
+        printf("Metering Mode: ");
         switch(ImageInfo.MeteringMode) {
-        case 2:
-            printf("Metering Mode: center weight\n");
-            break;
-        case 3:
-            printf("Metering Mode: spot\n");
-            break;
-        case 5:
-            printf("Metering Mode: matrix\n");
-            break;
+        case 1: printf("average\n"); break;
+        case 2: printf("center weight\n"); break;
+        case 3: printf("spot\n"); break;
+        case 4: printf("multi spot\n");  break;
+        case 5: printf("pattern\n"); break;
+        case 6: printf("partial\n");  break;
+        case 255: printf("other\n");  break;
+        default: printf("unknown (%d)\n",ImageInfo.MeteringMode); break;
         }
     }
 
@@ -1501,7 +1508,7 @@ void ShowImageInfo(int ShowFileInfo)
     if (ImageInfo.Process != M_SOF0){
         // don't show it if its the plain old boring 'baseline' process, but do
         // show it if its something else, like 'progressive' (used on web sometimes)
-        int a;
+        unsigned a;
         for (a=0;;a++){
             if (a >= PROCESS_TABLE_SIZE){
                 // ran off the end of the table.
@@ -1521,11 +1528,15 @@ void ShowImageInfo(int ShowFileInfo)
         if (ImageInfo.GpsAlt[0]) printf("GPS Altitude : %s\n",ImageInfo.GpsAlt);
     }
 
+    if (ImageInfo.QualityGuess){
+        printf("JPEG Quality : %d\n", ImageInfo.QualityGuess);
+    }
+
     // Print the comment. Print 'Comment:' for each new line of comment.
     if (ImageInfo.Comments[0]){
         int a,c;
         printf("Comment      : ");
-        if (!ImageInfo.CommentWidchars){
+        if (!ImageInfo.CommentWidthchars){
             for (a=0;a<MAX_COMMENT_SIZE;a++){
                 c = ImageInfo.Comments[a];
                 if (c == '\0') break;
@@ -1542,7 +1553,7 @@ void ShowImageInfo(int ShowFileInfo)
             }
             printf("\n");
         }else{
-            printf("%.*ls\n", ImageInfo.CommentWidchars, (wchar_t *)ImageInfo.Comments);
+            printf("%.*ls\n", ImageInfo.CommentWidthchars, (wchar_t *)ImageInfo.Comments);
         }
     }
 }
