@@ -2,12 +2,12 @@
 // Program to pull the information out of various types of EXIF digital 
 // camera files and show it in a reasonably consistent way
 //
-// Version 2.97
+// Version 3.00
 //
 // Compiling under Windows:  
 //   Make sure you have Microsoft's compiler on the path, then run make.bat
 //
-// Dec 1999 - Jan 2013
+// Dec 1999 - Feb 2015
 //
 // by Matthias Wandel   www.sentex.net/~mwandel
 //--------------------------------------------------------------------------
@@ -19,7 +19,7 @@
 
 #include <sys/stat.h>
 
-#define JHEAD_VERSION "2.97"
+#define JHEAD_VERSION "3.00"
 
 // This #define turns on features that are too very specific to 
 // how I organize my photos.  Best to ignore everything inside #ifdef MATTHIAS
@@ -56,6 +56,7 @@ static int Quiet        = FALSE;    // Be quiet on success (like unix programs)
        int DumpExifMap  = FALSE;
 static int ShowConcise  = FALSE;
 static int CreateExifSection = FALSE;
+static int TrimExifTrailingZeroes = FALSE;
 static char * ApplyCommand = NULL;  // Apply this command to all images.
 static char * FilterModel = NULL;
 static int    ExifOnly    = FALSE;
@@ -178,7 +179,7 @@ static int FileEditComment(char * TempFileName, char * Comment, int CommentSize)
     }
 
     // Read the file back in.
-    CommentSize = fread(Comment, 1, 999, file);
+    CommentSize = fread(Comment, 1, MAX_COMMENT_SIZE, file);
 
     fclose(file);
 
@@ -243,6 +244,7 @@ static int ModifyDescriptComment(char * OutComment, char * SrcComment)
                                 if (!memcmp(Line, AddComment, l+1)){
                                     TagExists = TRUE;
                                     strncpy(Line, AddComment, sizeof(Line));
+                                    Line[sizeof(Line)-1]='\0';
                                     Modified = TRUE;
                                 }
                             }
@@ -282,21 +284,27 @@ static int AutoResizeCmdStuff(void)
 {
     static char CommandString[PATH_MAX+1];
     double scale;
-    float TargetSize = 1600;
+    float TargetSize = 1800;
 
     ApplyCommand = CommandString;
 
     scale = TargetSize / ImageInfo.Width;
-    if (TargetSize / ImageInfo.Height > scale) scale = TargetSize  / ImageInfo.Width;
+    if (scale > TargetSize / ImageInfo.Height) scale = TargetSize / ImageInfo.Height;
 
     if (scale > 0.8){
+        if (ImageInfo.QualityGuess >= 93){
+            // Re-compress at lower quality.
+            sprintf(CommandString, "mogrify -quality 86 &i");
+            return TRUE;
+        }
         printf("not resizing %dx%x '%s'\n",ImageInfo.Height, ImageInfo.Width, ImageInfo.FileName);
         return FALSE;
     }
 
     if (scale < 0.4) scale = 0.4; // Don't scale down by too much.
 
-    sprintf(CommandString, "mogrify -geometry %dx%d -quality 80 &i",(int)(ImageInfo.Width*scale), (int)(ImageInfo.Height*scale));
+    sprintf(CommandString, "mogrify -geometry %dx%d -quality 85 &i",(int)(ImageInfo.Width*scale+0.5), 
+                                    (int)(ImageInfo.Height*scale+0.5));
     return TRUE;
 }
 
@@ -411,8 +419,21 @@ static void DoCommand(const char * FileName, int ShowIt)
         // Don't delete original file until we know a new one was created by the command.
         struct stat dummy;
         if (stat(TempName, &dummy) == 0){
+            struct stat buf;
+            int stat_result = stat(FileName, &buf);
+
             unlink(FileName);
             rename(TempName, FileName);
+            if (stat_result == 0){
+                // set Unix access rights and time to new file
+                struct utimbuf mtime;
+                chmod(FileName, buf.st_mode);
+
+                mtime.actime = buf.st_atime;
+                mtime.modtime = buf.st_mtime;
+            
+                utime(FileName, &mtime);
+            }
         }else{
             ErrFatal("specified command did not produce expected output file");
         }
@@ -539,8 +560,6 @@ void RenameAssociated(const char * FileName, char * NewBaseName)
 //--------------------------------------------------------------------------
 static void DoFileRenaming(const char * FileName)
 {
-    int NumAlpha = 0;
-    int NumDigit = 0;
     int PrefixPart = 0; // Where the actual filename starts.
     int ExtensionPart;  // Where the file extension starts.
     int a;
@@ -553,15 +572,10 @@ static void DoFileRenaming(const char * FileName)
     for (a=0;FileName[a];a++){
         if (FileName[a] == SLASH){
             // Don't count path component.
-            NumAlpha = 0;
-            NumDigit = 0;
             PrefixPart = a+1;
         }
 
         if (FileName[a] == '.') ExtensionPart = a;  // Remember where extension starts.
-
-        if (isalpha(FileName[a])) NumAlpha += 1;    // Tally up alpha vs. digits to judge wether to rename.
-        if (isdigit(FileName[a])) NumDigit += 1;
     }
 
     if (!Exif2tm(&tm, ImageInfo.DateTime)){
@@ -694,16 +708,16 @@ static void DoFileRenaming(const char * FileName)
 //--------------------------------------------------------------------------
 static int DoAutoRotate(const char * FileName)
 {
-    if (ImageInfo.Orientation >= 2 && ImageInfo.Orientation <= 8){
+    if (ImageInfo.Orientation != 1){
         const char * Argument;
         Argument = ClearOrientation();
 
         if (!ZeroRotateTagOnly){
             char RotateCommand[PATH_MAX*2+50];
             if (Argument == NULL){
-                ErrFatal("Orientation screwup");
+                ErrNonfatal("Unknown orientation tag",0,0);
+                return TRUE; // Image is still modified.
             }
-
             sprintf(RotateCommand, "jpegtran -trim -%s -outfile &o &i", Argument);
             ApplyCommand = RotateCommand;
             DoCommand(FileName, FALSE);
@@ -859,8 +873,9 @@ static void ProcessFile(const char * FileName)
             Modified = TRUE;
         }
         ReadMode = READ_IMAGE;   // Don't re-read exif section again on next read.
-
-    }else if (ExifXferScrFile){
+    }
+   
+    if (ExifXferScrFile){
         char RelativeExifName[PATH_MAX+1];
 
         // Make a relative name.
@@ -885,6 +900,35 @@ static void ProcessFile(const char * FileName)
         return;
     }
 
+    if (TrimExifTrailingZeroes){
+        if (ImageInfo.ThumbnailAtEnd){
+            Section_t * ExifSection;
+            int NumRedundant, WasRedundant;
+            unsigned char * StartRedundant;
+            //printf("Exif: Thumbnail %d - %d\n",ImageInfo.ThumbnailOffset, ImageInfo.ThumbnailOffset+ImageInfo.ThumbnailSize);
+            ExifSection = FindSection(M_EXIF);
+
+            StartRedundant = ExifSection->Data + 8 + ImageInfo.ThumbnailOffset+ImageInfo.ThumbnailSize;
+            WasRedundant = NumRedundant = (ExifSection->Size) - (ImageInfo.ThumbnailOffset + ImageInfo.ThumbnailSize + 8);
+            
+            //printf("Exif length: %d  Wasted: %d\n",ExifSection->Size, NumRedundant);
+
+            for(;NumRedundant > 0 && StartRedundant[NumRedundant-1] == 0;NumRedundant--);// Only remove trailing bytes if they are zero.
+
+            if (NumRedundant != WasRedundant){
+                int NewExifSize;
+                printf("Trimming %d bytes from exif in %s\n", WasRedundant-NumRedundant, FileName);
+                NewExifSize = ImageInfo.ThumbnailOffset + ImageInfo.ThumbnailSize + 8 + NumRedundant;
+                ExifSection->Data[0] = (uchar)(NewExifSize >> 8); // Must write new length into exif data.
+                ExifSection->Data[1] = (uchar)NewExifSize;
+                ExifSection->Size = NewExifSize;
+                Modified = TRUE;
+            }else{
+                //printf("Noting to remove from %s\n", FileName);
+            }
+        }
+    }
+
     FileSequence += 1; // Count files processed.
 
     if (ShowConcise){
@@ -985,7 +1029,7 @@ static void ProcessFile(const char * FileName)
             }else{
                 // Read it in.
                 // Replace the section.
-                CommentSize = fread(Comment, 1, 999, CommentFile);
+                CommentSize = fread(Comment, 1, MAX_COMMENT_SIZE, CommentFile);
                 fclose(CommentFile);
                 if (CommentSize < 0) CommentSize = 0;
             }
@@ -1044,9 +1088,8 @@ static void ProcessFile(const char * FileName)
             RelativeName(OutFileName, CommentSavefileName, FileName);
 
             CommentFile = fopen(OutFileName,"w");
-
             if (CommentFile){
-                fwrite((char *)CommentSec->Data+2, CommentSec->Size-2 ,1, CommentFile);
+                fwrite((char *)CommentSec->Data+2 ,CommentSec->Size-2, 1, CommentFile);
                 fclose(CommentFile);
             }else{
                 ErrFatal("Could not write comment file");
@@ -1233,6 +1276,7 @@ static void Usage (void)
            "  -ci <name> Insert comment section from a file.  -cs and -ci use same naming\n"
            "             scheme as used by the -st option\n"
            "  -cl string Insert literal comment string\n"
+           "  -zt        Trim exif header trailing zeroes (Nikon 1 wastes 30k that way)\n"
 
            "\nDATE / TIME MANIPULATION:\n"
            "  -ft        Set file modification time to Exif time\n"
@@ -1442,10 +1486,12 @@ int main (int argc, char **argv)
         }else if (!strcmp(arg,"-cl")){
             CommentInsertLiteral = argv[++argn];
             DoModify |= MODIFY_JPEG;
+        }else if (!strcmp(arg,"-zt")){
+            TrimExifTrailingZeroes = TRUE;
+            DoModify |= MODIFY_JPEG;
         }else if (!strcmp(arg,"-mkexif")){
             CreateExifSection = TRUE;
             DoModify |= MODIFY_JPEG;
-
     // Output verbosity control
         }else if (!strcmp(arg,"-h")){
             Usage();
